--╔═════════════════════════════╗--
--║ ┬┌─┌─┐┬ ┬┌┬┐┌─┐┌─┐┬─┐┌─┐┌─┐ ║--
--║ ├┴┐├┤ └┬┘│││├─┤│  ├┬┘│ │└─┐ ║--
--║ ┴ ┴└─┘ ┴ ┴ ┴┴ ┴└─┘┴└─└─┘└─┘ ║--
--╠═════════════════════════════╣--
--║   By Spyro [Mandokir EU]    ║--
--╚═════════════════════════════╝--

-- Upvalues
local _G, wipe, pairs, select, strlen, strmatch, C_Timer, PlaySound, CreateFrame, IsShiftKeyDown, IsAltKeyDown, IsControlKeyDown, UnitAffectingCombat, SetOverrideBindingClick, ClearOverrideBindings, GetSpecialization, GetSpecializationInfo, GetTalentTabInfo
    = _G, wipe, pairs, select, strlen, strmatch, C_Timer, PlaySound, CreateFrame, IsShiftKeyDown, IsAltKeyDown, IsControlKeyDown, UnitAffectingCombat, SetOverrideBindingClick, ClearOverrideBindings, GetSpecialization, GetSpecializationInfo, GetTalentTabInfo

-- Local namespace vars
local IsRetail = WOW_PROJECT_ID == WOW_PROJECT_MAINLINE
local _, DB -- Stores a local reference to the saved variables
local Event = CreateFrame("Frame") -- Frame for event processing
local Bindings = CreateFrame("Frame") -- Bindings owner to be able to clear all of them on spec change
local KeyCatcher = CreateFrame("Frame", "KeyMacros_KeyCatcher", UIParent) -- Will catch the pressed key to be binded
local Editor = CreateFrame("Frame", "KeyMacros_Editor", UIParent, "ButtonFrameTemplate") -- Macro editor window
local TotalMacroButtons = 0 -- Total of macro buttons created
local FreeMacroButtons = {} -- Stores the unused macro buttons, to be reused later
local KeyToButton = {} -- Stores each key binded and the name of its macro button that runs the macro. Needed when deleting binds
local IsModifierKey = { -- List of all modifiers
  [ "SHIFT"] = 1, [ "ALT"] = 1, [ "CTRL"] = 1,
  ["LSHIFT"] = 1, ["LALT"] = 1, ["LCTRL"] = 1,
  ["RSHIFT"] = 1, ["RALT"] = 1, ["RCTRL"] = 1,
}

-- GetNumSpecs()
-- Return the total numbers of specializations.
local GetNumSpecs = IsRetail and GetNumSpecializations or GetNumTalentTabs

-- GetSpecName()
-- Returns the name of a specialization.
local function GetSpecName(SpecIndex)
  if SpecIndex == 0 then return "Hybrid" -- Classic only
  elseif SpecIndex == 5 then return "None" -- Retail only (low level characters without a real spec)
  elseif IsRetail then return select(2, GetSpecializationInfo(SpecIndex)) -- Retail spec
  else return select(1, GetTalentTabInfo(SpecIndex)) end -- Classic talent tree
end

-- GetMacroButton()
-- Returns a MacroButton to be used in a SetOverrideBindingClick() call.
local function GetMacroButton()
  -- Checking if there is any unused button, to reuse it
  if FreeMacroButtons[#FreeMacroButtons] then
    local Button = FreeMacroButtons[#FreeMacroButtons]
    FreeMacroButtons[#FreeMacroButtons] = nil
    return Button
  end

  -- If there's none, create a new one
  TotalMacroButtons = TotalMacroButtons + 1
  local Button = CreateFrame("Button", "KeyMacros_MacroButton"..TotalMacroButtons, nil, "SecureActionButtonTemplate")
  Button:RegisterForClicks("AnyDown")
  Button:SetAttribute("type", "macro")

  return Button
end

-- FreeMacroButton()
-- Marks a macro button as free (unused), so it can be reused later.
local function FreeMacroButton(Button)
  FreeMacroButtons[#FreeMacroButtons + 1] = Button
  Button:SetAttribute("macrotext", nil)
end

-- CatchKey()
-- Captures the key pressed by the user to edit the macro that the key will run.
local function CatchKey()
  if not DB.CurrentSpec then
    print("|cFFFFFF00[KeyMacros]|r Select a specialization first by right-clicking on the minimap button.")
  elseif UnitAffectingCombat("player") then
    print("|cFFFFFF00[KeyMacros]|r Can't edit macros in combat!")
  else
    Editor:Hide()
    KeyCatcher:Show()
  end
end

-- UnbindKey()
-- Frees a key that is bound to execute a macrotext.
local function UnbindKey(Key)
  if not KeyToButton[Key] then return end -- Key is not bound
  SetOverrideBinding(Bindings, true, Key, nil)
  FreeMacroButton(KeyToButton[Key]) -- Free the button that this bind was using
  KeyToButton[Key] = nil
  DB[DB.CurrentSpec][Key] = nil
end

-- BindKey()
-- Makes a key execute macro code when pressed and saves it in the DB for the current spec.
local function BindKey(Key, Macro)
  if strlen(Macro) == 0 then -- If it's empty unbind it
    UnbindKey(Key)
    return
  end

  if KeyToButton[Key] then -- If it's already binded, just update the macro
    KeyToButton[Key]:SetAttribute("macrotext", Macro)
  else -- If not, create a new bind
    local Button = GetMacroButton()
    Button:SetAttribute("macrotext", Macro)
    SetOverrideBindingClick(Bindings, true, Key, Button:GetName()) -- Binding key to the macro
    KeyToButton[Key] = Button
  end

  DB[DB.CurrentSpec][Key] = Macro
end

-- LoadSpecBindings()
-- Loads the bindings (keys that execute macro code).
local function LoadSpecBindings(SpecIndex)
  ClearOverrideBindings(Bindings) -- Delete all the current bindings
  wipe(KeyToButton)

  -- Free all the currently used macro buttons
  local i = 1
  while _G["KeyMacros_MacroButton"..i] do
    FreeMacroButton(_G["KeyMacros_MacroButton"..i])
    i = i + 1
  end

  -- Loading the bindings for the spec
  for Key, Macro in pairs(DB[SpecIndex]) do BindKey(Key, Macro) end
end

-- DetectSpecData()
-- Waits for the spec info to be available and process all the spec-based data.
local function DetectSpecData()
  if not GetSpecName(IsRetail and GetSpecialization() or 1) then C_Timer.After(0.2, DetectSpecData) return end -- Waiting for spec info

  -- Adding specializations to the database
  for SpecIndex = IsRetail and 1 or 0, GetNumSpecs() do -- Classic starts from 0 coz that will be the "hybrid" spec
    DB[SpecIndex] = DB[SpecIndex] or {}
  end

  if IsRetail then
    DB.CurrentSpec = GetSpecialization() -- Saving current spec on Retail
    if DB.CurrentSpec == 5 then DB[5] = DB[5] or {} end -- Low level characters without a real spec

  -- Creating context menu to select the current spec in Classic (not needed in Retail)
  else
    local Menu = MSA_DropDownMenu_Create("KeyMacros_Menu", LibDBIcon10_KeyMacros)
    MSA_DropDownMenu_Initialize(Menu, function(self)
      local info = MSA_DropDownMenu_CreateInfo()

      info.text = "Specialization"
      info.isTitle = true
      info.notCheckable = true
      MSA_DropDownMenu_AddButton(info)
      info.isTitle = false
      info.notCheckable = false

      for SpecIndex = GetNumSpecs(), 0, -1 do
        info.text     = GetSpecName(SpecIndex)
        info.value    = SpecIndex
        info.checked  = info.value == DB.CurrentSpec
        info.disabled = false
        info.func     = function(self)
          MSA_DropDownMenu_SetSelectedValue(Menu, self.value, self.value)
          self.checked = true
          DB.CurrentSpec = self.value
          LoadSpecBindings(self.value)
        end

        MSA_DropDownMenu_AddButton(info)
      end
    end, "MENU")
  end

  if DB.CurrentSpec then LoadSpecBindings(DB.CurrentSpec) end
end

-- CancelStackSplit()
-- Closes the StackSplitFrame.
local function CancelStackSplit()
  StackSplitFrame:SetAlpha(1)
  if not StackSplitFrame:IsVisible() then return end
  (StackSplitCancelButton or StackSplitFrame.CancelButton):Click()
end

-- Hook for linking spells to the editor window
local SpellButton_ModifiedClick = function(self)
  local Slot = SpellBook_GetSpellBookSlot(self)
  if Slot > MAX_SPELLS or not IsModifiedClick("CHATLINK") or not Editor.EditBox:IsShown() or not Editor.EditBox:HasFocus() then return end

  local SpellName, SpellRank = GetSpellBookItemName(Slot, SpellBookFrame.bookType)
  if SpellName and not IsPassiveSpell(Slot, SpellBookFrame.bookType) then
    if SpellRank and strlen(SpellRank) > 0 then Editor.EditBox:Insert(SpellName.."("..SpellRank..")")
    else Editor.EditBox:Insert(SpellName) end
  end
end
if IsRetail then
  for i = 1, 12 do _G["SpellButton"..i]:HookScript("OnClick", SpellButton_ModifiedClick) end
else
  hooksecurefunc("SpellButton_OnModifiedClick", SpellButton_ModifiedClick)
end

-- Hook for linking items to the editor window
hooksecurefunc("HandleModifiedItemClick", function(Link)
  if not Link or not Editor.EditBox:IsShown() or not Editor.EditBox:HasFocus() then return end
  Editor.EditBox:Insert(strmatch(Link, "%[(.+)%]"))

  -- Cancelling the StackSplitFrame for splitable items
  StackSplitFrame:SetAlpha(0)
  C_Timer.After(0.01, CancelStackSplit)
end)

-- Frame to capture the key to bind to macro commands
KeyCatcher:SetSize(GetScreenWidth(), GetScreenHeight())
KeyCatcher:SetFrameStrata("TOOLTIP")
KeyCatcher:SetFrameLevel(10000)
KeyCatcher:SetPropagateKeyboardInput(false)
KeyCatcher:SetPoint("CENTER", UIParent)
KeyCatcher:Hide()

local KeyCatcherDialog = CreateFrame("Frame", "KeyMacros_KeyCatcherDialog", KeyCatcher, "DialogBoxFrame")
KeyMacros_KeyCatcherDialogButton:Hide()
KeyCatcherDialog.Center:SetColorTexture(0.15, 0.15, 0.14, 1)
KeyCatcherDialog:SetSize(558, 105)
KeyCatcherDialog:SetPoint("TOP", UIParent, "TOP", 0, -200)
KeyCatcherDialog:Show()

local PressKeyString = KeyCatcherDialog:CreateFontString()
PressKeyString:SetFont("Fonts/FRIZQT__.TTF", 20, "OUTLINE")
PressKeyString:SetPoint("CENTER")
PressKeyString:SetJustifyH("CENTER")
PressKeyString:SetText("Press the key you want to bind to macro commands\n\n|cFFFFD100ESC|r to cancel")

-- Function to process the key that have been pressed to be binded
local ShowEditor = function() Editor:Show() end -- To be used on C_Timer.After()
local KeyPressed = function(Key)
  if IsModifierKey[Key] then return end
  if Key == "ESCAPE" then KeyCatcher:Hide() return end

  -- Checking the modifiers pressed
  local Mod = ""
  if IsShiftKeyDown()   then Mod = Mod.."SHIFT-" end
  if IsAltKeyDown()     then Mod = Mod.."ALT-"   end
  if IsControlKeyDown() then Mod = Mod.."CTRL-"  end

  KeyCatcher:Hide()
  Editor.Key = Mod..Key
  C_Timer.After(0.01, ShowEditor) -- Needs to have a delay so the pressed key is not written in the macro editor
end

-- Intercepting pressed keys
KeyCatcher:SetScript("OnKeyDown", function(_, Key)
  KeyPressed(Key)
end)

-- Intercepting mouse buttons
KeyCatcher:SetScript("OnMouseDown", function(_, Button)
  local MouseKey = Button == "MiddleButton" and "BUTTON3" or Button == "Button4" and "BUTTON4" or Button == "Button5" and "BUTTON5"
  if MouseKey then KeyPressed(MouseKey) end
end)

-- Intercepting mousewheel
KeyCatcher:SetScript("OnMouseWheel", function(_, Delta)
  KeyPressed(Delta == 1 and "MOUSEWHEELUP" or "MOUSEWHEELDOWN")
end)

-- Editor window
KeyMacros_EditorTitleText:SetText("KeyMacros")
Editor:SetFrameStrata("DIALOG")
Editor:SetMovable(true)
Editor:SetSize(800, 500)
Editor:Hide()
Editor:SetPoint("CENTER", UIParent, "CENTER", 0, 0)
ButtonFrameTemplate_HidePortrait(Editor)
tinsert(UISpecialFrames, "KeyMacros_Editor")
if IsRetail then Editor.Bg:SetPoint("TOPLEFT", Editor, "TOPLEFT", 5, -21) end

local TitleBarMover = CreateFrame("Frame", nil, Editor)
TitleBarMover:SetPoint("TOPLEFT", Editor, "TOPLEFT", 0, 0)
TitleBarMover:SetSize(Editor:GetWidth() - 25, 21)
TitleBarMover:SetScript("OnMouseDown", function() Editor:StartMoving() end)
TitleBarMover:SetScript("OnMouseUp", function() Editor:StopMovingOrSizing() end)

local MacroForKeyLabel = Editor:CreateFontString()
MacroForKeyLabel:SetFont("Fonts/FRIZQT__.TTF", 20)
MacroForKeyLabel:SetPoint("TOPLEFT", IsRetail and 15 or 11, -33)
MacroForKeyLabel:SetJustifyH("LEFT")

local SpecNameLabel = Editor:CreateFontString()
SpecNameLabel:SetFont("Fonts/FRIZQT__.TTF", 20)
SpecNameLabel:SetPoint("TOPRIGHT", -11, -33)
SpecNameLabel:SetJustifyH("RIGHT")

local Scroll = CreateFrame("ScrollFrame", nil, Editor, "UIPanelScrollFrameTemplate")
Scroll:SetSize(Editor:GetWidth() - 50, Editor:GetHeight() - 102)
Scroll:SetPoint("TOPLEFT", 15, -69)

local EditBox = CreateFrame("EditBox", nil, Editor)
Editor.EditBox = EditBox
Scroll:SetScrollChild(EditBox)
EditBox:SetAutoFocus(false)
EditBox:SetMultiLine(true)
EditBox:SetFontObject(GameFontNormalLarge)
EditBox:SetTextColor(1, 1, 1)
EditBox:SetWidth(Scroll:GetWidth() - 10)
EditBox:SetScript("OnEscapePressed", function() EditBox:ClearFocus() end)

Editor:HookScript("OnShow", function()
  PlaySound(SOUNDKIT.IG_CHARACTER_INFO_OPEN)
  MacroForKeyLabel:SetText("|cFFFFD100Macro for key:|r "..Editor.Key)
  SpecNameLabel:SetText("|cFFFFD100Specialization:|r "..GetSpecName(DB.CurrentSpec))
  if DB[DB.CurrentSpec][Editor.Key] then EditBox:SetText(DB[DB.CurrentSpec][Editor.Key]) end
  EditBox:SetFocus(true)
end)

Editor:HookScript("OnHide", function()
  PlaySound(SOUNDKIT.IG_CHARACTER_INFO_CLOSE)
  EditBox:SetText("")
  EditBox:ClearFocus()
  Editor.Key = nil
end)

Editor:HookScript("OnMouseDown", function(self)
  EditBox:SetCursorPosition(strlen(EditBox:GetText()))
  EditBox:SetFocus(true)
end)

-- Buttons
Editor.SaveButton = CreateFrame("Button", nil, Editor, "GameMenuButtonTemplate")
Editor.SaveButton:SetText("Save")
Editor.SaveButton:SetPoint("BOTTOMLEFT", 4, 4)
Editor.SaveButton:SetScript("OnClick", function()
  BindKey(Editor.Key, EditBox:GetText())
  Editor:Hide()
end)

Editor.DeleteButton = CreateFrame("Button", nil, Editor, "GameMenuButtonTemplate")
Editor.DeleteButton:SetText("Delete")
Editor.DeleteButton:SetPoint("BOTTOM", 0, 4)
Editor.DeleteButton:SetScript("OnClick", function()
  if KeyToButton[Editor.Key] then UnbindKey(Editor.Key) end -- If it's bound, unbind it
  Editor:Hide()
end)

Editor.CancelButton = CreateFrame("Button", nil, Editor, "GameMenuButtonTemplate")
Editor.CancelButton:SetText("Cancel")
Editor.CancelButton:SetPoint("BOTTOMRIGHT", -4, 4)
Editor.CancelButton:SetScript("OnClick", function() Editor:Hide() end)

-- Character counter
Editor.CharCounter = Editor:CreateFontString() 
Editor.CharCounter:SetFont("Fonts/FRIZQT__.TTF", 11)
Editor.CharCounter:SetPoint("RIGHT", Editor.DeleteButton, "LEFT", -56, 0)
Editor.CharCounter:SetJustifyH("RIGHT")
EditBox:HookScript("OnTextChanged", function()
  local MacroLength = EditBox:GetNumLetters()
  if MacroLength > 1023 then Editor.CharCounter:SetText("|cFFFF0000"..MacroLength.."|r/1023") -- Red if we surpass the limit
  else Editor.CharCounter:SetText(MacroLength.."/1023") end
end)

-- LDB
local KeyMacrosLDB = LibStub("LibDataBroker-1.1"):NewDataObject("KeyMacros", {
  type = "launcher",
  text = "KeyMacros",
  icon = "Interface/MacroFrame/MacroFrame-Icon",

  OnEnter = function(Frame)
    GameTooltip:SetOwner(Frame, "ANCHOR_NONE")
    GameTooltip:SetPoint("RIGHT", Frame, "LEFT")
    GameTooltip:SetText("KeyMacros")
    GameTooltip:Show()
  end,

  OnLeave = function(Frame)
    GameTooltip:Hide()
  end,

  OnClick = function(_, Button)
    if Button == "LeftButton" then CatchKey()
    elseif Button == "MiddleButton" then CatchKey()
    elseif Button == "RightButton" then
      if IsRetail then
        CatchKey()
      else -- Spec selection menu for Classic
        Editor:Hide()
        GameTooltip:Hide()
        MSA_ToggleDropDownMenu(1, nil, KeyMacros_Menu, "cursor", 3, -3)
      end
    end
  end
})

-- Events
Event:SetScript("OnEvent", function(self, event, ...) return self[event](self, ...) end)
Event:RegisterEvent("ADDON_LOADED")
Event:RegisterEvent("PLAYER_REGEN_DISABLED")
if IsRetail then Event:RegisterEvent("PLAYER_SPECIALIZATION_CHANGED") end

-- Event ADDON_LOADED
-- Fires when an addon and its saved variables are loaded.
function Event:ADDON_LOADED(Name)
  if Name ~= "KeyMacros" then return end
  Event:UnregisterEvent("ADDON_LOADED")

  -- Creating SavedVariable
  KeyMacrosDB = KeyMacrosDB or {}
  DB = KeyMacrosDB

  -- Initializing minimap icon
  DB.minimap = DB.minimap or {}
  LibStub("LibDBIcon-1.0"):Register("KeyMacros", KeyMacrosLDB, DB.minimap)

  -- Creating a panel on Interface->Addons->KeyMacros
  local UI = CreateFrame("Frame", nil, InterfaceOptionsFrame)
  UI.name = "KeyMacros"
  InterfaceOptions_AddCategory(UI)

  -- Title
  UI.Title = UI:CreateFontString(nil, "ARTWORK", "GameFontNormalLarge")
  UI.Title:SetPoint("TOPLEFT", 16, -16)
  UI.Title:SetText("KeyMacros")

  -- Credits
  UI.Credits = UI:CreateFontString(nil, "ARTWORK", "GameFontWhite")
  UI.Credits:SetPoint("TOPRIGHT", -16, -16)
  UI.Credits:SetText("By |cFFFFD100Spyro|r [Mandokir EU]")

  -- Addon instructions
  UI.Tutorial = UI:CreateFontString(nil, "ARTWORK", "GameFontWhite")
  UI.Tutorial:SetPoint("TOPLEFT", UI.Title, "BOTTOMLEFT", -15, -10)
  UI.Tutorial:SetJustifyH("LEFT")
  UI.Tutorial:SetText([[
    To start using the addon left-click on the minimap button (or use the /keymacros command).
    You can also bind a key to execute /keymacros with the addon itself.

    All the macros are spec-specific, so you can have different macros for each spec.
    On Classic you will need to select your spec first by right-clicking on the minimap button
    (this is not needed on Retail).

    If you use KeyMacros on the vehicle keys (1, 2, 3, 4, 5, 6), you must add this at the start
    of your macro (with ActionButton1/2/3/4/5/6):
    |cFFFFD100/click [vehicleui][possessbar][overridebar] ActionButton1
    /stopmacro [vehicleui][possessbar][overridebar]
    <Rest of the macro here>
  ]])

  DetectSpecData()
end

-- Event PLAYER_REGEN_DISABLED
-- Fires when the player enters combat.
function Event:PLAYER_REGEN_DISABLED()
  if Editor:IsShown() then
    Editor:Hide() -- Hiding the editor coz we cannot edit macros in combat
    print("|cFFFFFF00[KeyMacros]|r Editing has been cancelled because you got in combat.")
  end
end

-- Event PLAYER_SPECIALIZATION_CHANGED
-- Triggers when a player changes his active spec. Retail only.
function Event:PLAYER_SPECIALIZATION_CHANGED(UnitID)
  if UnitID ~= "player" then return end

  local SpecIndex = GetSpecialization()
  if not SpecIndex then C_Timer.After(0.2, function() Event:PLAYER_SPECIALIZATION_CHANGED("player") end) return end -- Waiting for spec info

  if DB.CurrentSpec == 5 then -- Newly created character (spec 5) that didn't have a real spec until now
    for Key, Macro in pairs(DB[DB.CurrentSpec]) do -- Copying data from the temporal spec to the real one
      DB[SpecIndex][Key] = Macro
    end
  end

  DB.CurrentSpec = SpecIndex
  LoadSpecBindings(SpecIndex)
end

-- Command /keymacros
-- Starts the binding process.
SLASH_KEYMACROS1 = "/keymacros"
SlashCmdList["KEYMACROS"] = function()
  CatchKey()
end